import multiprocessing as mp
import os
from pathlib import Path
import subprocess

import ymmsl
from ymmsl import Operator, Port, Reference

from libmuscle.manager.manager import Manager
from libmuscle.manager.run_dir import RunDir

from .conftest import skip_if_python_only


def do_mmp_client_test(tmpdir, caplog):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )

    # create server
    ymmsl_doc = ymmsl.load(ymmsl_text)
    manager = Manager(ymmsl_doc, RunDir(Path(tmpdir)))

    # mock the deregistration
    removed_instance = None

    def mock_remove(name: Reference):
        nonlocal removed_instance
        removed_instance = name

    manager._instance_registry.remove = mock_remove

    # add some peers
    manager._instance_registry.add(
            Reference('macro'), ['tcp:test3', 'tcp:test4'],
            [Port('out', Operator.O_I), Port('in', Operator.S)])

    # create C++ client
    # it runs through the various RPC calls
    # see libmuscle/cpp/src/libmuscle/tests/mmp_client_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    env = os.environ.copy()
    lib_paths = [cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    if 'LD_LIBRARY_PATH' in env:
        env['LD_LIBRARY_PATH'] += ':' + ':'.join(map(str, lib_paths))
    else:
        env['LD_LIBRARY_PATH'] = ':'.join(map(str, lib_paths))

    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    cpp_test_client = cpp_test_dir / 'mmp_client_test'
    result = subprocess.run(
            [str(cpp_test_client), manager.get_server_location()], env=env)

    # check that C++-side checks were successful
    assert result.returncode == 0

    # check submit_log_message
    for rec in caplog.records:
        if rec.name == 'instances.test_logging':
            assert rec.time_stamp == '1970-01-01T00:00:02Z'
            assert rec.levelname == 'CRITICAL'
            assert rec.message == 'Integration testing'

    # check instance registry
    assert (manager._instance_registry.get_locations('micro[3]') ==
            ['tcp:test1', 'tcp:test2'])
    ports = manager._instance_registry.get_ports('micro[3]')
    assert ports[0].name == 'out'
    assert ports[0].operator == Operator.O_F
    assert ports[1].name == 'in'
    assert ports[1].operator == Operator.F_INIT

    # check deregister_instance
    assert removed_instance == 'micro[3]'

    manager.stop()


@skip_if_python_only
def test_mmp_client(log_file_in_tmpdir, tmpdir, caplog):
    process = mp.Process(target=do_mmp_client_test, args=(tmpdir, caplog))
    process.start()
    process.join()
    assert process.exitcode == 0



from libmuscle.manager.topology_store import TopologyStore

import pytest
from ymmsl import Configuration, Reference


def test_create_topology_store(topology_store) -> None:
    assert topology_store.conduits[0].sender == 'macro.out'
    assert topology_store.conduits[0].receiver == 'micro.in'
    assert topology_store.conduits[1].sender == 'micro.out'
    assert topology_store.conduits[1].receiver == 'macro.in'

    assert topology_store.kernel_dimensions['macro'] == []
    assert topology_store.kernel_dimensions['micro'] == [10, 10]


def test_get_conduits(topology_store2) -> None:
    conduits = topology_store2.get_conduits(Reference('macro'))
    assert conduits[0].sender == 'macro.out'
    assert conduits[0].receiver == 'meso.in'
    assert conduits[1].sender == 'meso.out'
    assert conduits[1].receiver == 'macro.in'

    conduits = topology_store2.get_conduits(Reference('meso'))
    assert conduits[0].sender == 'macro.out'
    assert conduits[0].receiver == 'meso.in'
    assert conduits[1].sender == 'meso.out'
    assert conduits[1].receiver == 'micro.in'
    assert conduits[2].sender == 'micro.out'
    assert conduits[2].receiver == 'meso.in'
    assert conduits[3].sender == 'meso.out'
    assert conduits[3].receiver == 'macro.in'


def test_get_peer_dimensions(topology_store) -> None:
    macro_peer_dims = topology_store.get_peer_dimensions(Reference('macro'))

    assert len(macro_peer_dims) == 1
    assert Reference('micro') in macro_peer_dims
    assert macro_peer_dims[Reference('micro')] == [10, 10]

    micro_peer_dims = topology_store.get_peer_dimensions(Reference('micro'))

    assert len(micro_peer_dims) == 1
    assert Reference('macro') in micro_peer_dims
    assert micro_peer_dims[Reference('macro')] == []


def test_data_error() -> None:
    config = Configuration('v0.1')
    with pytest.raises(ValueError):
        TopologyStore(config)



import logging
from pathlib import Path
from typing import cast, List, Optional

from ymmsl import Reference, Operator, Settings

from libmuscle.communicator import Communicator, Message
from libmuscle.mmp_client import MMPClient
from libmuscle.snapshot import MsgPackSnapshot, Snapshot, SnapshotMetadata

_logger = logging.getLogger(__name__)

_MAX_FILE_EXISTS_CHECK = 10000

# error text for save_snapshot when msg = None
_NO_MESSAGE_PROVIDED = (
        'Invalid message provided to `{}`. Please create a Message object to'
        ' store the state of the instance in a snapshot.')


class SnapshotManager:
    """Manages information on snapshots for the Instance

    Implements the public checkpointing API with handoffs to
    :class:`TriggerManager` for checkpoint triggers.
    """

    def __init__(self,
                 instance_id: Reference,
                 manager: MMPClient,
                 communicator: Communicator) -> None:
        """Create a new snapshot manager

        Args:
            instance_id: The id of this instance.
            manager: The client used to submit data to the manager.
            communicator: The communicator belonging to this instance.
        """
        self._instance_id = instance_id
        # replace identifier[i] by identifier-i to use in snapshot file name
        # using a dash (-) because that is not allowed in Identifiers
        self._safe_id = str(instance_id).replace("[", "-").replace("]", "")
        self._communicator = communicator
        self._manager = manager

        self._resume_from_snapshot = None   # type: Optional[Snapshot]
        self._resume_overlay = Settings()
        self._next_snapshot_num = 1

    def prepare_resume(
            self, resume_snapshot: Optional[Path],
            snapshot_directory: Optional[Path]) -> Optional[float]:
        """Apply checkpoint info received from the manager.

        If there is a snapshot to resume from, this loads it and does
        any resume work that libmuscle should do, including restoring
        message counts and storing the resumed-from snapshot again as
        our first snapshot.

        Args:
            resume_snapshot: Snapshot to resume from (or None if not
                resuming)
            snapshot_directory: directory to save snapshots in

        Returns:
            Time at which the initial snapshot was saved, if resuming.
        """
        result = None       # type: Optional[float]
        self._snapshot_directory = snapshot_directory or Path.cwd()
        if resume_snapshot is not None:
            snapshot = self.load_snapshot_from_file(resume_snapshot)

            if snapshot.message is not None:
                # snapshot.message is None for implicit snapshots
                self._resume_from_snapshot = snapshot
                result = snapshot.message.timestamp
            self._resume_overlay = snapshot.settings_overlay

            self._communicator.restore_message_counts(
                snapshot.port_message_counts)
            # Store a copy of the snapshot in the current run directory
            path = self.__store_snapshot(snapshot)
            metadata = SnapshotMetadata.from_snapshot(snapshot, str(path))
            self._manager.submit_snapshot_metadata(self._instance_id, metadata)

        return result

    def resuming_from_intermediate(self) -> bool:
        """Check whether we have an intermediate snapshot.

        Doesn't say whether we should resume now, just that we were
        given an intermediate snapshot to resume from by the manager.
        """
        return (
                self._resume_from_snapshot is not None and
                not self._resume_from_snapshot.is_final_snapshot)

    def resuming_from_final(self) -> bool:
        """Check whether we have a final snapshot.

        Doesn't say whether we should resume now, just that we were
        given an intermediate snapshot to resume from by the manager.
        """
        return (
                self._resume_from_snapshot is not None and
                self._resume_from_snapshot.is_final_snapshot)

    def load_snapshot(self) -> Message:
        """Get the Message to resume from.
        """
        snapshot = cast(Snapshot, self._resume_from_snapshot)
        return cast(Message, snapshot.message)

    def save_snapshot(
            self, msg: Optional[Message], final: bool,
            triggers: List[str], wallclock_time: float,
            f_init_max_timestamp: Optional[float],
            settings_overlay: Settings
            ) -> float:
        """Save a (final) snapshot.

        Args:
            msg: Message object representing the snapshot.
            final: True iff called from save_final_snapshot.
            triggers: Description of checkpoints that triggered this.
            wallclock_time: Wallclock time when saving.
            f_init_max_timestamp: Timestamp for final snapshots.

        Returns:
            Simulation time at which the snapshot was made
        """
        port_message_counts = self._communicator.get_message_counts()
        if final:
            # Decrease F_INIT port counts by one: F_INIT messages are already
            # pre-received, but not yet processed by the user code. Therefore,
            # the snapshot state should treat these as not-received.
            all_ports = self._communicator.list_ports()
            ports = all_ports.get(Operator.F_INIT, [])
            if self._communicator.settings_in_connected():
                ports.append('muscle_settings_in')
            for port_name in ports:
                new_counts = [i - 1 for i in port_message_counts[port_name]]
                port_message_counts[port_name] = new_counts

        snapshot = MsgPackSnapshot(
                triggers, wallclock_time, port_message_counts, final, msg,
                settings_overlay)

        path = self.__store_snapshot(snapshot)
        metadata = SnapshotMetadata.from_snapshot(snapshot, str(path))
        self._manager.submit_snapshot_metadata(self._instance_id, metadata)

        timestamp = msg.timestamp if msg is not None else float('-inf')
        if final and f_init_max_timestamp is not None:
            # For final snapshots f_init_max_snapshot is the reference time (see
            # should_save_final_snapshot).
            timestamp = f_init_max_timestamp
        return timestamp

    @staticmethod
    def load_snapshot_from_file(snapshot_location: Path) -> Snapshot:
        """Load a previously stored snapshot from the filesystem

        Args:
            snapshot_location: path where the snapshot is stored
        """
        _logger.debug(f'Loading snapshot from {snapshot_location}')
        if not snapshot_location.is_file():
            raise RuntimeError(f'Unable to load snapshot: {snapshot_location}'
                               ' is not a file. Please ensure this path exists'
                               ' and can be read.')

        # TODO: encapsulate I/O errors?
        with snapshot_location.open('rb') as snapshot_file:
            version = snapshot_file.read(1)
            data = snapshot_file.read()

            if version == MsgPackSnapshot.SNAPSHOT_VERSION_BYTE:
                return MsgPackSnapshot.from_bytes(data)
            raise RuntimeError('Unable to load snapshot from'
                               f' {snapshot_location}: unknown version of'
                               ' snapshot file. Was the file saved with a'
                               ' different version of libmuscle or'
                               ' edited?')

    def __store_snapshot(self, snapshot: Snapshot) -> Path:
        """Store a snapshot on the filesystem

        Args:
            snapshot: snapshot to store

        Returns:
            Path where the snapshot is stored
        """
        _logger.debug(f'Saving snapshot to {self._snapshot_directory}')
        for _ in range(_MAX_FILE_EXISTS_CHECK):
            # Expectation is that muscle_snapshot_directory is empty initially
            # and we succeed in the first loop. Still wrapping in a for-loop
            # such that an existing filename doesn't immediately raise an error
            fname = f"{self._safe_id}_{self._next_snapshot_num}.pack"
            fpath = self._snapshot_directory / fname
            self._next_snapshot_num += 1
            if not fpath.exists():
                break
        else:
            raise RuntimeError('Could not find an available filename for'
                               f' storing the next snapshot: {fpath} already'
                               ' exists.')
        # Opening with mode 'x' since a file with the same name may be created
        # in the small window between checking above and opening here. It is
        # better to fail with an error than to overwrite an existing file.
        with fpath.open('xb') as snapshot_file:
            snapshot_file.write(snapshot.SNAPSHOT_VERSION_BYTE)
            snapshot_file.write(snapshot.to_bytes())
        return fpath



from collections import OrderedDict

from ymmsl import (Component, Conduit, Configuration, Model, Operator,
                   Settings)

from libmuscle import Instance, Message, DONT_APPLY_OVERLAY
from libmuscle.runner import run_simulation


def qmc():
    """qMC implementation.
    """
    instance = Instance({Operator.O_F: ['settings_out[]']})

    while instance.reuse_instance():
        # o_f
        settings0 = Settings({'test2': 14.4})

        assert instance.is_connected('settings_out')
        assert instance.is_vector_port('settings_out')
        assert not instance.is_resizable('settings_out')
        length = instance.get_port_length('settings_out')
        assert length == 10
        for slot in range(length):
            instance.send('settings_out', Message(0.0, data=settings0), slot)


def macro():
    """Macro model implementation.
    """
    instance = Instance({
            Operator.O_I: ['out'], Operator.S: ['in']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test2') == 14.4
        # o_i
        instance.send('out', Message(0.0, 10.0, 'testing'))
        # s/b
        msg = instance.receive('in')
        assert msg.data == 'testing back'


def explicit_relay():
    """Intermediate component with explicit settings.

    Sends and receives overlay settings explicitly, rather than
    having MUSCLE handle them. This just passes all information on.
    """
    instance = Instance({
            Operator.F_INIT: ['in[]'], Operator.O_F: ['out[]']},
            DONT_APPLY_OVERLAY)

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test2', 'float') == 13.3
        assert instance.get_port_length('in') == instance.get_port_length(
                'out')

        msgs = list()
        for slot in range(instance.get_port_length('in')):
            msg = instance.receive_with_settings('in', slot)
            assert msg.data.startswith('testing')
            assert msg.settings['test2'] == 14.4
            msgs.append(msg)

        assert instance.get_setting('test2') == 13.3

        # o_f
        for slot in range(instance.get_port_length('out')):
            instance.send('out', msgs[slot], slot)


def micro():
    """Micro model implementation.
    """
    instance = Instance({Operator.F_INIT: ['in'], Operator.O_F: ['out']})

    assert instance.get_setting('test2') == 13.3
    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test2', 'float') == 14.4
        msg = instance.receive('in')
        assert msg.data == 'testing'

        # with pytest.raises(RuntimeError):
        #     instance.receive_with_settings('in')

        # o_f
        instance.send('out', Message(0.1, data='testing back'))


def test_settings_overlays(log_file_in_tmpdir):
    """A positive all-up test of settings overlays.
    """
    elements = [
            Component('qmc', 'qmc'),
            Component('macro', 'macro', [10]),
            Component('relay', 'explicit_relay'),
            Component('relay2', 'explicit_relay'),
            Component('micro', 'micro', [10])]

    conduits = [
                Conduit('qmc.settings_out', 'macro.muscle_settings_in'),
                Conduit('macro.out', 'relay.in'),
                Conduit('relay.out', 'micro.in'),
                Conduit('micro.out', 'relay2.in'),
                Conduit('relay2.out', 'macro.in')]

    model = Model('test_model', elements, conduits)

    settings = Settings(OrderedDict([
                ('test1', 13),
                ('test2', 13.3),
                ('test3', 'testing'),
                ('test4', True),
                ('test5', [2.3, 5.6]),
                ('test6', [[1.0, 2.0], [3.0, 1.0]])]))

    configuration = Configuration(model, settings)

    implementations = {
            'qmc': qmc,
            'macro': macro,
            'explicit_relay': explicit_relay,
            'micro': micro}

    run_simulation(configuration, implementations)



from pathlib import Path

import numpy as np

from libmuscle import Instance, Message
from ymmsl import Operator

from .conftest import skip_if_python_only, run_manager_with_actors


def macro():
    instance = Instance({
            Operator.O_I: ['out'],
            Operator.S: ['in']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test1') == 13

        for i in range(2):
            # o_i
            test_array = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
            assert test_array.shape == (2, 3)
            assert test_array.flags.c_contiguous
            data = {
                    'message': 'testing',
                    'test_grid': test_array}
            instance.send('out', Message(i * 10.0, (i + 1) * 10.0, data))

            # s/b
            msg = instance.receive('in')
            assert msg.data['reply'] == 'testing back {}'.format(i)
            assert msg.data['test_grid'].array.shape == (2, 3)
            assert msg.data['test_grid'].array.flags.f_contiguous
            assert (msg.data['test_grid'].array ==
                    np.array([[1, 2, 3], [4, 5, 6]])).all()
            assert msg.timestamp == i * 10.0


@skip_if_python_only
def test_fortran_macro_micro(mmp_server_config_simple, tmp_path):
    # create Fortran micro model
    # see libmuscle/fortran/src/libmuscle/tests/fortran_micro_model_test.f90
    run_manager_with_actors(
            mmp_server_config_simple,
            tmp_path,
            {},
            {'micro': Path('libmuscle') / 'tests' / 'fortran_micro_model_test'},
            {'macro': macro})



import socket
from typing import Optional

from libmuscle.mcp.transport_client import TransportClient
from libmuscle.mcp.tcp_util import recv_all, recv_int64, send_int64


class TcpTransportClient(TransportClient):
    """A client that connects to a TCPTransport server.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether this client class can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff this class can connect to this location.
        """
        return location.startswith('tcp:')

    def __init__(self, location: str) -> None:
        """Create a TcpClient for a given location.

        The client will connect to this location and be able to request
        messages from any instance and port represented by it.

        Args:
            location: A location string for the peer.
        """
        addresses = location[4:].split(',')

        sock = None     # type: Optional[socket.SocketType]
        for address in addresses:
            try:
                sock = self._connect(address)
                break
            except RuntimeError:
                pass

        if sock is None:
            raise RuntimeError('Could not connect to the server at location'
                               ' {}'.format(location))
        else:
            if hasattr(socket, "TCP_NODELAY"):
                sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
            if hasattr(socket, "TCP_QUICKACK"):
                sock.setsockopt(socket.SOL_TCP, socket.TCP_QUICKACK, 1)
            self._socket = sock

    def call(self, request: bytes) -> bytes:
        """Send a request to the server and receive the response.

        This is a blocking call.

        Args:
            request: The request to send

        Returns:
            The received response
        """
        send_int64(self._socket, len(request))
        self._socket.sendall(request)

        length = recv_int64(self._socket)
        return recv_all(self._socket, length)

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities.
        """
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()

    def _connect(self, address: str) -> socket.SocketType:
        loc_parts = address.rsplit(':', 1)
        host = loc_parts[0]
        if host.startswith('['):
            if host.endswith(']'):
                host = host[1:-1]
            else:
                raise RuntimeError('Invalid address')
        port = int(loc_parts[1])

        addrinfo = socket.getaddrinfo(
                host, port, 0, socket.SOCK_STREAM, socket.IPPROTO_TCP)

        for family, socktype, proto, _, sockaddr in addrinfo:
            try:
                sock = socket.socket(family, socktype, proto)
            except Exception:
                continue

            try:
                sock.connect(sockaddr)
            except Exception:
                sock.close()
                continue
            return sock

        raise RuntimeError('Could not connect')



import asyncio
import logging
import multiprocessing as mp
import os
from pathlib import Path
import queue
import sys
import traceback
from typing import Dict, List, Tuple

from qcg.pilotjob.allocation import (
        Allocation as qcg_Allocation, NodeAllocation as qcg_NodeAllocation)
from qcg.pilotjob.config import Config as qcg_Config
from qcg.pilotjob.errors import InternalError as qcg_InternalError
from qcg.pilotjob.executor import Executor as qcg_Executor
from qcg.pilotjob.joblist import (
        Job as qcg_Job, JobExecution as qcg_JobExecution,
        JobResources as qcg_JobResources)
from qcg.pilotjob.manager import (
        SchedulingIteration as qcg_SchedulingIteration,
        SchedulingJob as qcg_SchedulingJob)
from qcg.pilotjob.parseres import get_resources as qcg_get_resources
from qcg.pilotjob.resources import (
        Node as qcg_Node, ResourcesType as qcg_ResourcesType)

from ymmsl import ExecutionModel, MPICoresResReq, Reference, ThreadedResReq

from libmuscle.manager.instantiator import (
        CancelAllRequest, CrashedResult, InstantiationRequest, Process,
        ProcessStatus, QueueingLogHandler, ShutdownRequest)
from libmuscle.planner.planner import Resources


_logger = logging.getLogger(__name__)


class StateTracker:
    """Tracks processes and their state.

    This keeps a list of running processes and their state. It
    receives callbacks from QCG-PJ when the state of a job changes,
    and updates the list accordingly.

    Attributes:
        processes: Dict mapping instance names to Process objects.
    """
    def __init__(self) -> None:
        """Create a StateTracker."""
        self.processes = dict()     # type: Dict[Reference, Process]

        # These are for communicating with QCG-PJ
        self.queued_to_execute = 0
        self.stop_processing = False
        self.zmq_address = ''

    # QCG-PJ callbacks
    def job_executing(self, job_iteration: qcg_SchedulingIteration) -> None:
        """Called by Executor when a job has started to run.

        Args:
            job_iteration: The job iteration that has started
        """
        name = job_iteration._scheduling_job.job.name
        self.processes[name].status = ProcessStatus.RUNNING
        _logger.debug(f'Job {name} running')

    def job_finished(
            self, job_iteration: qcg_SchedulingIteration,
            allocation: qcg_Allocation,
            exit_code: int, error_msg: str, canceled: bool) -> None:
        """Called by executor when a job has finished.

        Args:
            job_iteration: The job iteration that has finished
            allocation: The allocation it ran on
            exit_code: Its exit code
            error_msg: Its error message
            canceled: Whether it was canceled
        """
        name = job_iteration._scheduling_job.job.name
        process = self.processes[name]
        if canceled:
            process.status = ProcessStatus.CANCELED
        elif exit_code == 0:
            process.status = ProcessStatus.SUCCESS
        else:
            process.status = ProcessStatus.ERROR

        process.exit_code = exit_code
        process.error_msg = error_msg


class QCGPJInstantiator(mp.Process):
    """Background process for interacting with the QCG-PJ executor."""
    def __init__(
            self, resources: mp.Queue, requests: mp.Queue, results: mp.Queue,
            log_records: mp.Queue, run_dir: Path) -> None:
        """Create a QCGPJProcessManager.

        Args:
            resources: Queue for returning the available resources
            requests: Queue to take requests from
            results: Queue to communicate finished processes over
            log_messages: Queue to push log messages to
        """
        super().__init__(name='QCGPJProcessManager')
        self._resources_out = resources
        self._requests_in = requests
        self._results_out = results
        self._log_records_out = log_records
        self._run_dir = run_dir

    def run(self) -> None:
        """Entry point for the process."""
        # Put QCG-PJ output in run dir
        # The configuration setting below is ignored by the agents
        # due to a bug in QCG-PJ
        qcgpj_dir = self._run_dir / 'qcgpj'
        qcgpj_dir.mkdir(exist_ok=True)
        os.chdir(qcgpj_dir)

        self._reconfigure_logging()

        # Executor needs to be instantiated before we go async
        qcg_config = {
                qcg_Config.AUX_DIR: str(qcgpj_dir)}     # type: Dict[str, str]
        self._qcg_resources = qcg_get_resources(qcg_config)
        self._state_tracker = StateTracker()
        self._executor = qcg_Executor(
                self._state_tracker, qcg_config, self._qcg_resources)

        self._send_resources()

        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self._main())
        except:     # noqa
            for line in traceback.format_exception(*sys.exc_info()):
                _logger.error(line)
            self._results_out.put(CrashedResult())

    async def _main(self) -> None:
        """Main function for the background process.

        This sets up QCG-PJ and then accepts requests for instantiating
        jobs, stopping them, or shutting down. Results of finished jobs
        are returned via the results queue.
        """
        qcg_iters = dict()  # type: Dict[Reference, qcg_SchedulingIteration]

        await asyncio.sleep(0.01)  # allow requests_in queue to be populated

        shutting_down = False
        done = False
        while not done:
            while not shutting_down:
                try:
                    request = self._requests_in.get_nowait()
                    if isinstance(request, ShutdownRequest):
                        _logger.debug('Got ShutdownRequest')
                        self._state_tracker.stop_processing = True
                        shutting_down = True

                    elif isinstance(request, CancelAllRequest):
                        _logger.debug('Got CancelAllRequest')
                        await self._cancel_all(qcg_iters)
                        _logger.debug('Done CancelAllRequest')

                    elif isinstance(request, InstantiationRequest):
                        if not shutting_down:
                            qcg_alloc, qcg_iter = self._create_job(
                                    request, self._qcg_resources.rtype)
                            qcg_iters[request.instance] = qcg_iter
                            self._state_tracker.processes[request.instance] = (
                                    Process(
                                        request.instance, request.resources))
                            self._state_tracker.queued_to_execute += 1
                            await self._executor.execute(qcg_alloc, qcg_iter)

                except queue.Empty:
                    break

            for name, process in list(self._state_tracker.processes.items()):
                if process.status.is_finished():
                    _logger.debug(f'Reporting {name} done')
                    self._results_out.put(process)
                    del self._state_tracker.processes[name]

            if shutting_down:
                _logger.debug(f'Done: {self._state_tracker.processes}')
                done = len(self._state_tracker.processes) == 0

            if not done:
                await asyncio.sleep(0.1)

        _logger.debug('Stopping executor')
        await self._executor.stop()

    def _reconfigure_logging(self) -> None:
        """Reconfigure logging to send to log_records_out."""
        root_logger = logging.getLogger()
        for h in list(root_logger.handlers):
            root_logger.removeHandler(h)

        handler = QueueingLogHandler(self._log_records_out)
        root_logger.addHandler(handler)

    def _send_resources(self) -> None:
        """Converts and sends QCG available resources."""
        resources = Resources()
        for node in self._qcg_resources.nodes:
            resources.cores[node.name] = set(map(int, node.free_ids))

        self._resources_out.put(resources)

    async def _cancel_all(
            self, qcg_iters: Dict[Reference, qcg_SchedulingIteration]) -> None:
        """Cancels all running jobs."""
        # Repeat cancel until they're gone to work around QCG-PJ
        # race condition.
        while any([
                not p.status.is_finished()
                for p in self._state_tracker.processes.values()]):

            for instance, process in self._state_tracker.processes.items():
                if process.status.is_finished():
                    continue

                qcg_iter = qcg_iters[instance]
                try:
                    await self._executor.cancel_iteration(
                            qcg_iter.job, qcg_iter.iteration)
                    _logger.debug(f'Canceled {instance}')
                except qcg_InternalError:
                    _logger.debug(f'Canceled {instance} not found')
                    raise
                except AttributeError:
                    # Workaround for QCG-PJ bug
                    _logger.debug(f'Canceled {instance} not found')
                    raise
            await asyncio.sleep(0.1)

    def _create_job(
            self, request: InstantiationRequest,
            qcg_resources_type: qcg_ResourcesType
            ) -> Tuple[qcg_Allocation, qcg_SchedulingIteration]:
        """Creates a QCG allocation and job for a request."""
        total_cores = sum(map(len, request.resources.cores.values()))

        env = self._create_env(request.instance, request.implementation.env)

        if request.implementation.script:
            execution = self._qcg_job_execution_with_script(request, env)
        else:
            execution = self._qcg_job_execution_normal(
                    request, env, qcg_resources_type)

        resources = qcg_JobResources(numCores=total_cores)

        qcg_job = qcg_Job(
                str(request.instance),
                execution=execution,
                resources=resources)

        qcg_allocation = qcg_Allocation()
        for node_name, cores in request.resources.cores.items():
            qcg_cores = [str(i) for i in cores]
            qcg_allocation.add_node(
                    qcg_NodeAllocation(qcg_Node(node_name), qcg_cores, {}))

        sjob = qcg_SchedulingJob(self._state_tracker, qcg_job)
        qcg_iteration = qcg_SchedulingIteration(sjob, None, None, None, [])
        return qcg_allocation, qcg_iteration

    def _create_env(
            self, instance: Reference, overlay: Dict[str, str]
            ) -> Dict[str, str]:
        """Updates the environment with the implementation's env.

        This updates env in-place. Keys from overlay that start with
        + will have the corresponding value appended to the matching
        (by key, without the +) value in env, otherwise the value in
        env gets overwritten.
        """
        env = os.environ.copy()
        env['MUSCLE_INSTANCE'] = str(instance)

        for key, value in overlay.items():
            if key.startswith('+'):
                if key[1:] in env:
                    env[key[1:]] += value
                else:
                    env[key[1:]] = value
            else:
                env[key] = value
        return env

    def _qcg_job_execution_with_script(
            self, request: InstantiationRequest, env: Dict[str, str]
            ) -> qcg_JobExecution:
        """Create a JobExecution with a run script."""
        impl = request.implementation
        if impl.script is None:
            raise RuntimeError()

        script_file = request.instance_dir / 'run_script.sh'
        with script_file.open('w') as f:
            f.write(impl.script)
        script_file.chmod(0o700)

        if isinstance(request.res_req, ThreadedResReq):
            env['MUSCLE_THREADS'] = str(request.res_req.threads)
        elif isinstance(request.res_req, MPICoresResReq):
            # OpenMPI support
            rank_file = request.instance_dir / 'rankfile'
            with rank_file.open('w') as f:
                i = 0
                for node, cores in request.resources.cores.items():
                    for c in sorted(cores):
                        f.write(f'rank {i}={node} slot={c}\n')
                        i += 1
            env['MUSCLE_OPENMPI_RANK_FILE'] = str(rank_file)

            # IntelMPI support
            mpi_res_args = list()
            for node, cores in request.resources.cores.items():
                mpi_res_args.extend(['-host', node, '-n', str(len(cores))])
            env['MUSCLE_INTELMPI_RESOURCES'] = ' '.join(mpi_res_args)

            # General environment
            env['MUSCLE_MPI_PROCESSES'] = str(
                    request.res_req.mpi_processes)
            env['MUSCLE_THREADS_PER_MPI_PROCESS'] = str(
                    request.res_req.threads_per_mpi_process)

        return qcg_JobExecution(
                exec=str(script_file),
                env=env,
                stdout=str(request.stdout_path),
                stderr=str(request.stderr_path),
                wd=str(request.work_dir),
                model='default')

    def _qcg_job_execution_normal(
            self, request: InstantiationRequest, env: Dict[str, str],
            qcg_resources_type: qcg_ResourcesType) -> qcg_JobExecution:
        """Create a JobExecution for a normal description."""
        impl = request.implementation
        total_cores = sum(map(len, request.resources.cores.values()))

        if impl.execution_model == ExecutionModel.DIRECT:
            env['OMP_NUM_THREADS'] = str(total_cores)
        else:
            env['OMP_NUM_THREADS'] = '1'

        model_map = {
                ExecutionModel.DIRECT: 'threads',
                ExecutionModel.OPENMPI: 'openmpi',
                ExecutionModel.INTELMPI: 'intelmpi',
                ExecutionModel.SRUNMPI: 'srunmpi'}
        qcg_execution_model = model_map[impl.execution_model]

        executable = str(impl.executable)
        args = impl.args if impl.args is not None else []
        if qcg_resources_type == qcg_ResourcesType.LOCAL:
            if impl.execution_model == ExecutionModel.DIRECT:
                if not impl.virtual_env and not impl.modules:
                    # QCG-PJ uses a bash call like this in all but this
                    # case, so we add it here to be consistent.
                    cmd = ' '.join([executable] + args)
                    executable = 'bash'
                    args = ['-l', '-c', cmd]
            elif impl.execution_model == ExecutionModel.OPENMPI:
                executable, args = self._with_local_open_mpi(
                        executable, args, total_cores)
            elif impl.execution_model == ExecutionModel.INTELMPI:
                executable, args = self._with_local_intel_mpi(
                        executable, args, total_cores)
            elif impl.execution_model == ExecutionModel.SRUNMPI:
                raise RuntimeError(
                    f'Cannot instantiate implementation {impl.name} with'
                    ' execution model "srunmpi", because there is no'
                    ' Slurm/srun here.')

        _logger.debug(f'Starting {executable} with {args}')

        return qcg_JobExecution(
                exec=executable,
                args=args,
                env=env,
                stdout=str(request.stdout_path),
                stderr=str(request.stderr_path),
                modules=impl.modules,
                venv=str(impl.virtual_env) if impl.virtual_env else None,
                wd=str(request.work_dir),
                model=qcg_execution_model,
                model_opts={'srun_opts': ['--overlap']})

    def _with_local_open_mpi(
            self, executable: str, args: List[str], num_processes: int
            ) -> Tuple[str, List[str]]:
        """Create OpenMPI mpirun call."""
        new_args = ['-n', str(num_processes), executable]
        new_args.extend(args)
        return 'mpirun', new_args

    def _with_local_intel_mpi(
            self, executable: str, args: List[str], num_processes: int
            ) -> Tuple[str, List[str]]:
        """Create IntelMPI mpirun call."""
        new_args = ['-n', str(num_processes), executable]
        new_args.extend(args)
        return 'mpirun', new_args



from enum import Enum
import time
from typing import Optional

from ymmsl import Port, Reference

from libmuscle.timestamp import Timestamp


class ProfileEventType(Enum):
    """Profiling event types for MUSCLE3.

    These match the types in the MUSCLE Manager Protocol, and should \
    be kept identical to those.
    """
    REGISTER = 0
    CONNECT = 4
    DEREGISTER = 1
    SEND = 2
    RECEIVE = 3


class ProfileEvent:
    """A profile event as used by MUSCLE3.

    This represents a single measurement of the timing of some event
    that occurred while executing the simulation.

    Args:
        instance_id: The identifier of the instance that generated \
                this message.
        start_time: When the event started (real-world, not \
                simulation time).
        stop_time: When the event ended (real-world, not simulation \
                time).
        event_type: Type of event that was measured.
        port: Port used for sending or receiving, if applicable.
        port_length: Length of that port, if a vector.
        slot: Slot that was sent or received on, if applicable.
        message_size: Size of the message involved, if applicable.

    Attributes:
        instance_id: The identifier of the instance that generated \
                this message.
        start_time: When the event started (real-world, not \
                simulation time).
        stop_time: When the event ended (real-world, not simulation \
                time).
        event_type: Type of event that was measured.
        port: Port used for sending or receiving, if applicable.
        port_length: Length of that port, if a vector.
        slot: Slot that was sent or received on, if applicable.
        message_size: Size of the message involved, if applicable.
    """
    def __init__(
            self,
            instance_id: Reference,
            start_time: Timestamp,
            stop_time: Timestamp,
            event_type: ProfileEventType,
            port: Optional[Port] = None,
            port_length: Optional[int] = None,
            slot: Optional[int] = None,
            message_size: Optional[int] = None
            ) -> None:

        self.instance_id = instance_id
        self.start_time = start_time
        self.stop_time = stop_time
        self.event_type = event_type
        self.port = port
        self.port_length = port_length
        self.slot = slot
        self.message_size = message_size

    def stop(self) -> None:
        """Sets stop_time to the current time.
        """
        self.stop_time = Timestamp(time.time())



from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from ymmsl import (
        Configuration, Model, Component, Conduit, Implementation,
        KeepsStateForNextUse, Reference)

from libmuscle.manager.snapshot_registry import (
    SnapshotNode, SnapshotRegistry, calc_consistency, calc_consistency_list,
    safe_get, _ConnectionInfo)
from libmuscle.manager.topology_store import TopologyStore
from libmuscle.snapshot import SnapshotMetadata


def make_snapshot(**msg_counts) -> SnapshotMetadata:
    return SnapshotMetadata([], 0, 0, 0, {**msg_counts}, False, '')


@pytest.fixture(params=[True, False])
def micro_is_stateless(request: pytest.FixtureRequest) -> bool:
    return request.param


@pytest.fixture
def macro_micro(micro_is_stateless: bool) -> Configuration:
    components = [
            Component('macro', 'macro_impl'),
            Component('micro', 'micro_impl')]
    conduits = [
            Conduit('macro.o_i', 'micro.f_i'),
            Conduit('micro.o_f', 'macro.s')]
    model = Model('macro_micro', components, conduits)

    if micro_is_stateless:
        micro_impl = Implementation(
                'micro_impl',
                keeps_state_for_next_use=KeepsStateForNextUse.NO,
                executable='pass')
    else:
        micro_impl = Implementation('micro_impl', executable='pass')

    implementations = [
            Implementation('macro_impl', executable='pass'),
            micro_impl]

    return Configuration(model, implementations=implementations)


@pytest.fixture
def uq(macro_micro: Configuration) -> Configuration:
    for component in macro_micro.model.components:
        component.multiplicity = [5]
    macro_micro.model.components.append(Component('qmc', 'qmc_impl'))
    macro_micro.model.components.append(Component('rr', 'rr_impl'))
    macro_micro.model.conduits.extend([
            Conduit('qmc.parameters_out', 'rr.front_in'),
            Conduit('rr.front_out', 'qmc.states_in'),
            Conduit('rr.back_out', 'macro.muscle_settings_in'),
            Conduit('macro.final_state_out', 'rr.back_in')])
    macro_micro.implementations[Reference('qmc_impl')] = Implementation(
            'qmc_impl', executable='pass')
    macro_micro.implementations[Reference('rr_impl')] = Implementation(
            'rr_impl', executable='pass')
    return macro_micro


def test_safe_get() -> None:
    assert safe_get([], 0, 1) == 1
    assert safe_get([3], 0, 1) == 3
    assert safe_get([3], 1, 5) == 5
    for i in range(10):
        expected = -1 if i >= 3 else i + 3
        assert safe_get([3, 4, 5], i, -1) == expected


def test_calc_consistency() -> None:
    num_sent = 3
    for num_received in [2, 3, 4, 5]:
        expect = num_received in [3, 4]
        assert calc_consistency(num_sent, num_received, True, False) is expect
        assert calc_consistency(num_received, num_sent, False, False) is expect

    num_received = 10
    for num_sent in [8, 9, 10, 11]:
        expect = num_sent in [9, 10]
        assert calc_consistency(num_sent, num_received, True, False) is expect
        assert calc_consistency(num_received, num_sent, False, False) is expect


def test_calc_consistency_with_restart() -> None:
    # Check normal rules
    assert calc_consistency(0, 0, True, True)
    assert calc_consistency(0, 0, False, True)
    assert not calc_consistency(1, 0, True, True)
    assert not calc_consistency(1, 0, True, False)
    assert calc_consistency(1, 0, False, False)
    # Different: num2 == 0 comes from the restarted actor, we do not want a
    # resume file to be created in this instance (because an instance further in
    # the call chain is ahead of the one that would be restarted):
    assert not calc_consistency(1, 0, False, True)


def test_calc_consistency_list() -> None:
    num_sent = [3, 3]
    for num_received in [[2, 3], [3, 2], [3, 5], [], [4, 4, 0, 0, 2]]:
        assert not calc_consistency_list(num_sent, num_received, True, False)
        assert not calc_consistency_list(num_received, num_sent, False, False)
    for num_received in [[3, 3], [3, 4], [4, 3], [4, 4],
                         [3, 3, 1], [4, 4, 0, 0, 0, 1, 0, 1]]:
        assert calc_consistency_list(num_sent, num_received, True, False)
        assert calc_consistency_list(num_received, num_sent, False, False)


def test_write_ymmsl(tmp_path: Path):
    configuration = Configuration(Model('empty', []))
    snapshot_registry = SnapshotRegistry(
            configuration, tmp_path, TopologyStore(configuration))
    snapshot_registry._write_snapshot_ymmsl([])

    paths = list(tmp_path.iterdir())
    assert len(paths) == 1
    assert paths[0].suffix == ".ymmsl"
    paths[0].unlink()

    now = datetime.now()
    for seconds in range(3):
        time = (now + timedelta(seconds=seconds)).strftime("%Y%m%d_%H%M%S")
        (tmp_path / f'snapshot_{time}.ymmsl').touch()
    snapshot_registry._write_snapshot_ymmsl([])
    paths = list(tmp_path.iterdir())
    assert len(paths) == 4
    paths = list(tmp_path.glob('*_1.ymmsl'))
    assert len(paths) == 1


def test_snapshot_config():
    configuration = Configuration(Model('empty', []))
    snapshot_registry = SnapshotRegistry(
            configuration, None, TopologyStore(configuration))
    micro_metadata = SnapshotMetadata(
            ['simulation_time >= 24.0', 'wallclocktime >= 10'],
            10.123456789, 24.3456789, None, {}, False, 'micro_snapshot')
    macro_metadata = SnapshotMetadata(
            ['simulation_time >= 12.0', 'wallclocktime >= 10'],
            10.123456789, 12.3456789, None, {}, False, 'macro_snapshot')
    snapshots = [
            SnapshotNode(1, Reference('micro'), micro_metadata, set()),
            SnapshotNode(1, Reference('macro'), macro_metadata, set())]

    now = datetime.now()
    config = snapshot_registry._generate_snapshot_config(snapshots, now)
    assert len(config.resume) == 2
    assert config.resume[Reference('macro')] == Path('macro_snapshot')
    assert config.resume[Reference('micro')] == Path('micro_snapshot')
    # note: no automatic testing for formatting, should verify by eye if this
    # looks okay..
    print(config.description)

    long_metadata = SnapshotMetadata(
            ['simulation_time >= 24.0'], 1.23456789e-10, 1.23456789e10, None,
            {}, False, '/this/is/a/long/path/to/the/snapshot/file.pack')
    snapshots.append(SnapshotNode(
            1, Reference('this.is.a.long.reference[10]'), long_metadata, set()))

    config = snapshot_registry._generate_snapshot_config(snapshots, now)
    assert len(config.resume) == 3
    assert config.resume[Reference('this.is.a.long.reference[10]')] == Path(
            '/this/is/a/long/path/to/the/snapshot/file.pack')
    print(config.description)


def test_peers(uq: Configuration) -> None:
    snapshot_registry = SnapshotRegistry(uq, None, TopologyStore(uq))
    macro = Reference('macro')
    micro = Reference('micro')
    qmc = Reference('qmc')
    rr = Reference('rr')

    all_instances = {qmc, rr} | {macro + i for i in range(5)}
    all_instances.update(micro + i for i in range(5))
    assert snapshot_registry._instances == all_instances

    assert snapshot_registry._get_peers(qmc) == {rr}
    expected_rr_peers = {qmc} | {macro + i for i in range(5)}
    assert snapshot_registry._get_peers(rr) == expected_rr_peers
    for i in range(5):
        assert snapshot_registry._get_peers(macro + i) == {rr, micro + i}
        assert snapshot_registry._get_peers(micro + i) == {macro + i}


def test_connections(uq: Configuration) -> None:
    snapshot_registry = SnapshotRegistry(uq, None, TopologyStore(uq))
    macro = Reference('macro')
    micro = Reference('micro')
    qmc = Reference('qmc')
    rr = Reference('rr')

    assert not snapshot_registry._get_connections(qmc, macro + 1)
    assert not snapshot_registry._get_connections(macro + 3, qmc)
    assert not snapshot_registry._get_connections(qmc, micro + 0)
    assert not snapshot_registry._get_connections(micro + 1, qmc)
    assert not snapshot_registry._get_connections(rr, micro + 4)
    assert not snapshot_registry._get_connections(micro + 0, rr)

    connections = snapshot_registry._get_connections(rr, qmc)
    assert len(connections) == 2
    for rr_port, qmc_port, info in connections:
        assert rr_port in (Reference('front_out'), Reference('front_in'))
        assert qmc_port in (Reference('parameters_out'), Reference('states_in'))
        is_sending = bool(info & _ConnectionInfo.SELF_IS_SENDING)
        assert is_sending is (rr_port == Reference('front_out'))
        # Note: actually both are vector ports, but this is undetectable from
        # the ymmsl configuration. Luckily we treat it the same as scalar-scalar
        assert not (info & _ConnectionInfo.SELF_IS_VECTOR)
        assert not (info & _ConnectionInfo.PEER_IS_VECTOR)

    connections = snapshot_registry._get_connections(macro + 0, rr)
    assert len(connections) == 2
    for macro_port, rr_port, info in connections:
        assert macro_port in (
                Reference('muscle_settings_in'), Reference('final_state_out'))
        assert rr_port in (Reference('back_out'), Reference('back_in'))
        is_sending = bool(info & _ConnectionInfo.SELF_IS_SENDING)
        assert is_sending is (macro_port == Reference('final_state_out'))
        assert not (info & _ConnectionInfo.SELF_IS_VECTOR)
        assert (info & _ConnectionInfo.PEER_IS_VECTOR)

    connections = snapshot_registry._get_connections(rr, macro + 1)
    assert len(connections) == 2
    for rr_port, macro_port, info in connections:
        assert macro_port in (
                Reference('muscle_settings_in'), Reference('final_state_out'))
        assert rr_port in (Reference('back_out'), Reference('back_in'))
        is_sending = bool(info & _ConnectionInfo.SELF_IS_SENDING)
        assert is_sending is (rr_port == Reference('back_out'))
        assert (info & _ConnectionInfo.SELF_IS_VECTOR)
        assert not (info & _ConnectionInfo.PEER_IS_VECTOR)


def test_implementation(uq: Configuration) -> None:
    snapshot_registry = SnapshotRegistry(uq, None, TopologyStore(uq))

    qmc_impl = snapshot_registry._implementation(Reference('qmc'))
    assert qmc_impl.name == 'qmc_impl'

    missing_impl = snapshot_registry._implementation(Reference('missing'))
    assert missing_impl is None


def test_macro_micro_snapshots(macro_micro: Configuration) -> None:
    snapshot_registry = SnapshotRegistry(
            macro_micro, None, TopologyStore(macro_micro))
    # prevent actually writing a ymmsl file, testing that separately
    snapshot_registry._write_snapshot_ymmsl = MagicMock()
    macro = Reference('macro')
    micro = Reference('micro')

    macro_snapshot = make_snapshot(o_i=[3], s=[3])
    snapshot_registry._add_snapshot(macro, macro_snapshot)

    assert len(snapshot_registry._snapshots[macro]) == 1
    node = snapshot_registry._snapshots[macro][0]
    assert node.consistent is False
    assert node.consistent_peers == {}
    assert node.instance == macro
    assert node.num == 1
    assert node.snapshot is macro_snapshot
    assert node.peers == {micro}
    snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    # Note: this snapshot is not realistic, it should have come in before
    # the macro snapshot above. However, it's still useful for testing the
    # consistency algorithm
    micro_snapshot = make_snapshot(f_i=[2], o_f=[1])
    snapshot_registry._add_snapshot(micro, micro_snapshot)

    assert len(snapshot_registry._snapshots[micro]) == 1
    assert snapshot_registry._snapshots[micro][0].consistent is False
    snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    micro_snapshot = make_snapshot(f_i=[3], o_f=[2])
    snapshot_registry._add_snapshot(micro, micro_snapshot)

    # The first micro snapshots should be cleaned up now
    assert len(snapshot_registry._snapshots[micro]) == 1
    micro_node = snapshot_registry._snapshots[micro][0]
    assert micro_node.consistent
    snapshot_registry._write_snapshot_ymmsl.assert_called_once_with(
            [micro_node, node])
    snapshot_registry._write_snapshot_ymmsl.reset_mock()

    # 3 micro snapshots in the same reuse:
    for _ in range(3):
        micro_snapshot = make_snapshot(f_i=[4], o_f=[3])
        snapshot_registry._add_snapshot(micro, micro_snapshot)

    # Previous micro snapshot should be cleaned up now
    assert len(snapshot_registry._snapshots[micro]) == 1
    micro_node = snapshot_registry._snapshots[micro][-1]
    assert snapshot_registry._write_snapshot_ymmsl.call_count == 3
    snapshot_registry._write_snapshot_ymmsl.assert_called_with(
            [micro_node, node])
    snapshot_registry._write_snapshot_ymmsl.reset_mock()

    macro_snapshot = make_snapshot(o_i=[4], s=[4])
    snapshot_registry._add_snapshot(macro, macro_snapshot)
    snapshot_registry._write_snapshot_ymmsl.assert_called_once()
    snapshot_registry._write_snapshot_ymmsl.reset_mock()

    # 3 micro snapshots in the same reuse, but inconcistent with previous macro
    for _ in range(3):
        micro_snapshot = make_snapshot(f_i=[6], o_f=[5])
        snapshot_registry._add_snapshot(micro, micro_snapshot)

    # All three should be present now in addition to the one last used in
    # workflow snapshot
    assert len(snapshot_registry._snapshots[micro]) == 4
    snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    macro_snapshot = make_snapshot(o_i=[6], s=[6])
    snapshot_registry._add_snapshot(macro, macro_snapshot)
    assert snapshot_registry._write_snapshot_ymmsl.call_count == 3
    assert len(snapshot_registry._snapshots[micro]) == 1
    assert len(snapshot_registry._snapshots[macro]) == 1


def test_uq(uq: Configuration) -> None:
    snapshot_registry = SnapshotRegistry(uq, None, TopologyStore(uq))
    # prevent actually writing a ymmsl file, testing that separately
    snapshot_registry._write_snapshot_ymmsl = MagicMock()
    macro = Reference('macro')
    micro = Reference('micro')
    qmc = Reference('qmc')
    rr = Reference('rr')

    qmc_snapshot = make_snapshot(parameters_out=[], states_in=[])
    snapshot_registry._add_snapshot(qmc, qmc_snapshot)

    rr_snapshot = make_snapshot(
            front_in=[1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
            front_out=[0] * 10,
            back_out=[1, 1, 1, 1, 1],
            back_in=[0] * 5)
    snapshot_registry._add_snapshot(rr, rr_snapshot)
    node = snapshot_registry._snapshots[rr][-1]
    assert qmc in node.consistent_peers
    snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    macro_snapshot = make_snapshot(
            muscle_settings_in=[1], final_state_out=[0], o_i=[0], s=[0])
    for i in range(5):
        snapshot_registry._add_snapshot(macro + i, macro_snapshot)
        node = snapshot_registry._snapshots[macro + i][-1]
        assert node.consistent_peers.keys() == {rr}
        snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    micro_snapshot = make_snapshot(f_i=[1], o_f=[0])
    for i in range(5):
        snapshot_registry._add_snapshot(micro + i, micro_snapshot)
        node = snapshot_registry._snapshots[micro + i][-1]
        assert node.consistent_peers.keys() == {macro + i}
        if i == 4:
            snapshot_registry._write_snapshot_ymmsl.assert_called_once()
            snapshot_registry._write_snapshot_ymmsl.reset_mock()
        else:
            snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    qmc_snapshot = make_snapshot(parameters_out=[1, 1, 1, 1, 1], states_in=[])
    snapshot_registry._add_snapshot(qmc, qmc_snapshot)
    node = snapshot_registry._snapshots[qmc][-1]
    assert node.consistent_peers.keys() == {rr}
    snapshot_registry._write_snapshot_ymmsl.assert_called_once()
    snapshot_registry._write_snapshot_ymmsl.reset_mock()
    assert len(snapshot_registry._snapshots[qmc]) == 1  # previous is cleaned up


def test_heuristic_rollbacks() -> None:
    components = [Component(f'comp{i}', f'impl{i}') for i in range(4)]
    conduits = [Conduit(f'comp{i}.o_f', f'comp{i+1}.f_i') for i in range(3)]
    model = Model('linear', components, conduits)
    implementations = [
            Implementation(f'impl{i}', script='xyz')
            for i in range(4)]
    config = Configuration(model, implementations=implementations)

    comp1, comp2, comp3, comp4 = (Reference(f'comp{i}') for i in range(4))

    snapshot_registry = SnapshotRegistry(config, None, TopologyStore(config))
    # prevent actually writing a ymmsl file, testing that separately
    snapshot_registry._write_snapshot_ymmsl = MagicMock()

    for i in range(4):
        snapshot_registry._add_snapshot(comp1, make_snapshot(o_f=[i]))
    assert len(snapshot_registry._snapshots[comp1]) == 4

    for i in range(10):
        snapshot_registry._add_snapshot(
                comp2, make_snapshot(f_i=[1], o_f=[0]))
        snapshot_registry._add_snapshot(
                comp3, make_snapshot(f_i=[1], o_f=[0]))
    assert len(snapshot_registry._snapshots[comp2]) == 10
    assert len(snapshot_registry._snapshots[comp3]) == 10

    snapshot_registry._add_snapshot(comp2, make_snapshot(f_i=[2], o_f=[1]))
    assert len(snapshot_registry._snapshots[comp2]) == 11
    snapshot_registry._add_snapshot(comp2, make_snapshot(f_i=[3], o_f=[2]))
    assert len(snapshot_registry._snapshots[comp2]) == 12

    snapshot_registry._write_snapshot_ymmsl.assert_not_called()

    snapshot_registry._add_snapshot(
            comp4, make_snapshot(f_i=[1]))
    snapshot_registry._write_snapshot_ymmsl.assert_called()

    assert len(snapshot_registry._snapshots[comp1]) == 2
    assert len(snapshot_registry._snapshots[comp2]) == 2
    assert len(snapshot_registry._snapshots[comp3]) == 1
    assert len(snapshot_registry._snapshots[comp4]) == 1



from copy import copy, deepcopy
import logging
from typing import Dict, Iterable, List, Mapping, Optional, Set

from ymmsl import (
        Component, Configuration, Model, MPICoresResReq, MPINodesResReq,
        Operator, Reference, ResourceRequirements, ThreadedResReq)


_logger = logging.getLogger(__name__)


_PredSuccType = Dict[Component, Set[Component]]


class ModelGraph:
    """Represents a yMMSL model as a graph."""
    def __init__(self, model: Model) -> None:
        """Create a ModelGraph.

        This is essentially a helper class that makes it easier to
        analyse a yMMSL model definition.

        Args:
            model: The model to represent.
        """
        self._model = model

        self._direct_superpreds = dict()    # type: _PredSuccType
        self._direct_predecessors = dict()  # type: _PredSuccType
        self._direct_subpreds = dict()      # type: _PredSuccType

        self._direct_supersuccs = dict()    # type: _PredSuccType
        self._direct_successors = dict()    # type: _PredSuccType
        self._direct_subsuccs = dict()      # type: _PredSuccType

        self._superpreds = dict()           # type: _PredSuccType
        self._predecessors = dict()         # type: _PredSuccType
        self._subpreds = dict()             # type: _PredSuccType

        self._supersuccs = dict()           # type: _PredSuccType
        self._successors = dict()           # type: _PredSuccType
        self._subsuccs = dict()             # type: _PredSuccType

        self._calc_direct_succs_preds()
        self._calc_predecessors()
        self._calc_successors()

    def components(self) -> Iterable[Component]:
        """Return the components of the model (nodes)."""
        return self._model.components

    def component(self, name: Reference) -> Component:
        """Return a component by name.

        Args:
            name: Name of the component to look up

        Returns:
            The component with the given name

        Raises:
            KeyError: If no component could be found
        """
        for component in self._model.components:
            if component.name == name:
                return component
        raise KeyError('Component {} not found'.format(name))

    def successors(self, component: Component) -> Set[Component]:
        """Return the successors of the given component.

        Args:
            component: The reference component

        Returns:
            The set of components for which a path exists from
            component's O_F to their F_INIT.

        Raises:
            KeyError: If the component is not in the model used to
                construct this object.
        """
        return self._successors[component]

    def predecessors(self, component: Component) -> Set[Component]:
        """Return the predecessors of the given component.

        Args:
            component: The reference component

        Returns:
            The set of components for which a path exists from
            their O_F to component's F_INIT.

        Raises:
            KeyError: If the component is not in the model used to
                construct this object.
        """
        return self._predecessors[component]

    def macros(self, component: Component) -> Set[Component]:
        """Return the macros of the given component.

        These are components that are both before the component's
        F_INIT and after its O_F.

        Args:
            component: The reference component

        Returns:
            The set of components that are both super-predecessor
            and super-successor of component.
        """
        return self._superpreds[component] & self._supersuccs[component]

    def micros(self, component: Component) -> Set[Component]:
        """Return the micros of the given component.

        These are components that are in between the component's
        O_I and its S.

        Args:
            component: The reference component

        Returns:
            The set of components that are both sub-successor
            and sub-predecessor of component.
        """
        return self._subsuccs[component] & self._subpreds[component]

    def _calc_predecessors(self) -> None:
        """Calculates predecessors of each component in the model.

        This function determines for each component in the model which
        other components can be reached by following incoming conduits.

        Preconditions:
            self._model set
            self._direct_* calculated

        Side effects:
            Fills self._superpreds, self._predecessors, self._subpreds
            with for each component in the model the corresponding set
            of components.
        """
        self._superpreds = {c: set() for c in self._model.components}
        self._predecessors = {c: set() for c in self._model.components}
        self._subpreds = {c: set() for c in self._model.components}

        remaining_preds = {
                c: (
                    len(self._direct_predecessors[c]) +
                    len(self._direct_superpreds[c]))
                for c in self._model.components}
        remaining_subpreds = {
                c: len(self._direct_subpreds[c])
                for c in self._model.components}

        todo = set(self._model.components)
        started = set()     # type: Set[Component]
        doing = set()       # type: Set[Component]
        finished = set()    # type: Set[Component]
        done = set()        # type: Set[Component]
        while todo or doing:
            started.clear()
            for component in todo:
                if not remaining_preds[component]:
                    for subsucc in self._direct_subsuccs[component]:
                        self._superpreds[subsucc].add(component)
                        self._predecessors[subsucc].update(
                                self._predecessors[component])
                        self._superpreds[subsucc].update(
                                self._superpreds[component])
                        remaining_preds[subsucc] -= 1
                    started.add(component)

            todo -= started
            doing |= started

            finished.clear()
            for component in doing:
                if not remaining_subpreds[component]:
                    for succ in self._direct_successors[component]:
                        self._predecessors[succ].add(component)
                        self._predecessors[succ].update(
                                self._subpreds[component])
                        self._predecessors[succ].update(
                                self._predecessors[component])
                        self._superpreds[succ].update(
                                self._superpreds[component])
                        remaining_preds[succ] -= 1

                    for supersucc in self._direct_supersuccs[component]:
                        self._subpreds[supersucc].add(component)
                        self._subpreds[supersucc].update(
                                self._subpreds[component])
                        self._subpreds[supersucc].update(
                                self._predecessors[component])
                        remaining_subpreds[supersucc] -= 1

                    finished.add(component)

            doing -= finished
            done |= finished

    def _calc_successors(self) -> None:
        """Calculates successors of each component in the model.

        This function determines for each component in the model which
        other components can be reached by following outgoing conduits.

        Preconditions:
            self._model set
            self._direct_* calculated

        Side effects:
            Fills self._supersuccs, self._successors, self._subsuccs
            with for each component in the model the corresponding set
            of components.
        """
        self._supersuccs = {c: set() for c in self._model.components}
        self._successors = {c: set() for c in self._model.components}
        self._subsuccs = {c: set() for c in self._model.components}

        remaining_succs = {
                c: (
                    len(self._direct_successors[c]) +
                    len(self._direct_supersuccs[c]))
                for c in self._model.components}

        remaining_subsuccs = {
                c: len(self._direct_subsuccs[c])
                for c in self._model.components}

        todo = set(self._model.components)
        started = set()     # type: Set[Component]
        doing = set()       # type: Set[Component]
        finished = set()    # type: Set[Component]
        done = set()        # type: Set[Component]
        while todo or doing:
            started.clear()
            for component in todo:
                if not remaining_succs[component]:
                    for subpred in self._direct_subpreds[component]:
                        self._supersuccs[subpred].add(component)
                        self._successors[subpred].update(
                                self._successors[component])
                        self._supersuccs[subpred].update(
                                self._supersuccs[component])
                        remaining_succs[subpred] -= 1
                    started.add(component)

            todo -= started
            doing |= started

            finished.clear()
            for component in doing:
                if not remaining_subsuccs[component]:
                    for pred in self._direct_predecessors[component]:
                        self._successors[pred].add(component)
                        self._successors[pred].update(
                                self._successors[component])
                        self._supersuccs[pred].update(
                                self._supersuccs[component])
                        self._successors[pred].update(
                                self._subsuccs[component])
                        remaining_succs[pred] -= 1

                    for superpred in self._direct_superpreds[component]:
                        self._subsuccs[superpred].add(component)
                        self._subsuccs[superpred].update(
                                self._successors[component])
                        self._subsuccs[superpred].update(
                                self._subsuccs[component])
                        remaining_subsuccs[superpred] -= 1
                    finished.add(component)

            doing -= finished
            done |= finished

    def _calc_direct_succs_preds(self) -> None:
        """Calculates all successors and predecessors of components.

        Preconditions:
            self._model set

        Side effects:
            Sets self._direct_successors to a dictionary mapping each
            component in the model to the set of components that it
            has a * -> F_INIT conduit to.
        """
        components = {c.name: c for c in self._model.components}
        self._direct_supersuccs = {c: set() for c in self._model.components}
        self._direct_successors = {c: set() for c in self._model.components}
        self._direct_subsuccs = {c: set() for c in self._model.components}
        self._direct_superpreds = {c: set() for c in self._model.components}
        self._direct_predecessors = {c: set() for c in self._model.components}
        self._direct_subpreds = {c: set() for c in self._model.components}

        for conduit in self._model.conduits:
            sender = components[conduit.sending_component()]
            snd_op = None
            if sender.ports:
                if conduit.sending_port() in sender.ports.o_i:
                    snd_op = Operator.O_I
                elif conduit.sending_port() in sender.ports.o_f:
                    snd_op = Operator.O_F

            receiver = components[conduit.receiving_component()]
            recv_op = None
            if conduit.receiving_port() == 'muscle_settings_in':
                recv_op = Operator.F_INIT
            elif receiver.ports:
                if conduit.receiving_port() in receiver.ports.f_init:
                    recv_op = Operator.F_INIT
                elif conduit.receiving_port() in receiver.ports.s:
                    recv_op = Operator.S

            if (snd_op, recv_op) == (Operator.O_I, Operator.F_INIT):
                self._direct_superpreds[receiver].add(sender)
                self._direct_subsuccs[sender].add(receiver)
            elif (snd_op, recv_op) == (Operator.O_F, Operator.F_INIT):
                self._direct_successors[sender].add(receiver)
                self._direct_predecessors[receiver].add(sender)
            elif (snd_op, recv_op) == (Operator.O_F, Operator.S):
                self._direct_subpreds[receiver].add(sender)
                self._direct_supersuccs[sender].add(receiver)


class Resources:
    """Designates a (sub)set of resources.

    Whether these resources are free or allocated in general or by
    something specific depends on the context, this just says which
    resources we're talking about.

    Attributes:
        cores: A dictionary mapping designated nodes to designated
                cores on them.
    """
    def __init__(self, cores: Optional[Dict[str, Set[int]]] = None) -> None:
        """Create a Resources object with the given cores.

        Args:
            cores: Cores to be designated by this object.
        """
        if cores is None:
            self.cores = dict()     # type: Dict[str, Set[int]]
        else:
            self.cores = cores

    def __copy__(self) -> 'Resources':
        """Copy the object."""
        return Resources(deepcopy(self.cores))

    def __iadd__(self, other: 'Resources') -> 'Resources':
        """Add the resources in the argument to this object."""
        for node in other.cores:
            if node in self.cores:
                self.cores[node] |= other.cores[node]
            else:
                self.cores[node] = set(other.cores[node])
        return self

    def __isub__(self, other: 'Resources') -> 'Resources':
        """Remove the resources in the argument from this object."""
        for node in other.cores:
            if node in self.cores:
                self.cores[node] -= other.cores[node]
                if not self.cores[node]:
                    del self.cores[node]
        return self

    def __str__(self) -> str:
        """Return a human-readable string representation."""
        return 'Resources(' + '; '.join([
            n + ': ' + ','.join(sorted(map(str, cs)))
            for n, cs in self.cores.items()]) + ')'

    def __repr__(self) -> str:
        """Return a string representation."""
        return f'Resources({self.cores})'

    def nodes(self) -> Iterable[str]:
        """Returns the nodes on which we designate resources."""
        return self.cores.keys()

    def total_cores(self) -> int:
        """Returns the total number of cores designated."""
        return sum([len(cs) for cs in self.cores.values()])

    def isdisjoint(self, other: 'Resources') -> bool:
        """Returns whether we share resources with other."""
        for node, cores in self.cores.items():
            if node in other.cores:
                if not cores.isdisjoint(other.cores[node]):
                    return False
        return True

    @staticmethod
    def union(resources: Iterable['Resources']) -> 'Resources':
        """Combines the resources into one.

        Args:
            resources: A collection of resources to merge.

        Return:
            A Resources object referring to all the resources in the
            input.
        """
        result = Resources()
        for cur_resources in resources:
            result += cur_resources
        return result


class InsufficientResourcesAvailable(RuntimeError):
    pass


class Planner:
    """Allocates resources and keeps track of allocations."""
    def __init__(self, all_resources: Resources):
        """Create a ResourceManager.

        Args:
            all_resources: An object describing the available resources
                    to be managed by this ResourceManager.
        """
        self._all_resources = all_resources
        self._allocations = dict()  # type: Dict[Reference, Resources]
        self._oversubscribed = dict()   # type: Dict[Reference, Resources]
        self._next_virtual_node = 1

    def allocate_all(
            self, configuration: Configuration, virtual: bool = False
            ) -> Dict[Reference, Resources]:
        """Allocates resources for the given components.

        Allocation can occur either on a fixed set of available
        resources (virtual set to  False), or on an elastic set of
        virtual resources (virtual set to True). Use the former
        inside of a (fixed) cluster allocation and the latter to
        determine how many nodes are needed to run a simulation
        without oversubscribing.

        If virtual is True, additional nodes will be added as
        needed to obtain the resources needed to allocate all
        instances. Each additional node will have as many cores
        as a random existing one. The intended use is to pass a
        single node to __init__ when using this mode.

        Args:
            configuration: Configuration to allocate all components of
            virtual: Allocate on virtual resources or not, see above

        Returns:
            Resources for each instance required by the model.
        """
        result = dict()     # type: Dict[Reference, Resources]

        # Analyse model
        model = ModelGraph(configuration.model)
        requirements = configuration.resources
        implementations = configuration.implementations
        exclusive = {
                c for c in model.components()
                if (c.implementation and
                    not implementations[c.implementation].can_share_resources)}

        # Allocate
        unallocated_instances = [
                i for c in model.components() for i in c.instances()]
        leftover_instances = list()     # type: List[Reference]
        while unallocated_instances:
            leftover_instances.clear()

            to_allocate = self._sort_instances(
                    unallocated_instances, requirements)

            for instance in to_allocate:
                component = model.component(instance.without_trailing_ints())
                conflicting_names = self._conflicting_names(
                        model, exclusive, component)

                done = False
                while not done:
                    try:
                        result[instance] = self._allocate_instance(
                                instance, component,
                                requirements[component.name],
                                conflicting_names, virtual)
                        done = True
                    except InsufficientResourcesAvailable:
                        if virtual:
                            self._expand_resources(
                                    component.name,
                                    requirements[component.name])
                        else:
                            leftover_instances.append(instance)
                            done = True

            if leftover_instances:
                _logger.warning(
                    'Planner ran out of resources, oversubscribing remaining'
                    ' instances.')
                self._oversubscribed.update(self._allocations)
                self._allocations.clear()

            unallocated_instances.clear()
            unallocated_instances.extend(leftover_instances)

        return result

    def _sort_instances(
            self, instances: List[Reference],
            requirements: Mapping[Reference, ResourceRequirements]
            ) -> List[Reference]:
        """Return to be allocated components in optimal order.

        This is a heuristic, it's not actually optimal but it should
        give decent results most of the time.

        Args:
            instances: The instances to sort
            requirements: The resource requirements for their
                components, indexed by name
        """
        cmp_names = map(Reference.without_trailing_ints, instances)
        reqs = map(lambda n: requirements[n], cmp_names)
        instances_reqs = list(zip(instances, reqs))
        threaded = [
                (i, r.threads) for i, r in instances_reqs
                if isinstance(r, ThreadedResReq)]
        sorted_threaded = sorted(threaded, key=lambda x: x[1], reverse=True)
        sorted_threaded_instances = [x[0] for x in sorted_threaded]

        mpi = [
                (i, r.mpi_processes) for i, r in instances_reqs
                if isinstance(r, MPICoresResReq)]
        sorted_mpi = sorted(mpi, key=lambda x: x[1], reverse=True)
        sorted_mpi_instances = [x[0] for x in sorted_mpi]

        return sorted_threaded_instances + sorted_mpi_instances

    def _conflicting_names(
            self, model: ModelGraph, exclusive: Set[Component],
            component: Component) -> Set[Reference]:
        """Returns names of components that cannot share resources."""
        conflicting_comps = set(model.components())
        conflicting_comps -= model.predecessors(component)
        conflicting_comps -= model.successors(component)
        if component not in exclusive:
            mms = model.micros(component) | model.macros(component)
            nonconflicting_mms = mms - exclusive
            conflicting_comps -= nonconflicting_mms
        return {c.name for c in conflicting_comps}

    def _expand_resources(
            self, name: Reference, req: ResourceRequirements) -> None:
        """Adds an extra virtual node to the available resources."""
        taken = True
        while taken:
            new_node = 'node{:06d}'.format(self._next_virtual_node)
            taken = new_node in self._all_resources.cores
            self._next_virtual_node += 1

        num_cores = len(next(iter(self._all_resources.cores.values())))
        if isinstance(req, ThreadedResReq):
            if req.threads > num_cores:
                raise InsufficientResourcesAvailable(
                        f'Instance {name} requires {req.threads} threads,'
                        f' which is impossible with {num_cores} cores per'
                        ' node.')
        if isinstance(req, MPICoresResReq):
            if req.threads_per_mpi_process > num_cores:
                raise InsufficientResourcesAvailable(
                        f'Instance {name} requires'
                        f' {req.threads_per_mpi_process} threads per process,'
                        f' which is impossible with {num_cores} cores per'
                        ' node.')
        self._all_resources.cores[new_node] = set(range(num_cores))

    def _allocate_instance(
            self, instance: Reference, component: Component,
            requirements: ResourceRequirements,
            simultaneous_components: Set[Reference], virtual: bool
            ) -> Resources:
        """Allocates resources for the given instance.

        If we are on real resources, and the instance requires more
        threads than our nodes have cores, then we'll just give it all
        cores on a node and hope for the best. If we are on virtual
        resources, this will raise InsufficientResourcesAvailable.

        Args:
            instance: The instance to allocate for
            component: The component it is an instance of
            requirements: Its resource requirements
            simultaneous_components: Components which may execute
                    simultaneously and whose resources we therefore
                    cannot overlap with
            virtual: Whether we are on virtual resources

        Returns:
            A Resources object describing the resources allocated
        """
        allocation = Resources({})
        free_resources = copy(self._all_resources)

        for other in self._allocations:
            if other.without_trailing_ints() in simultaneous_components:
                free_resources -= self._allocations[other]

        try:
            if isinstance(requirements, ThreadedResReq):
                allocation = self._allocate_thread_block(
                        free_resources, requirements.threads)

            elif isinstance(requirements, MPICoresResReq):
                if requirements.threads_per_mpi_process != 1:
                    raise RuntimeError(
                            'Multiple threads per MPI process is not supported'
                            ' yet. Please make an issue on GitHub.')
                for proc in range(requirements.mpi_processes):
                    allocation += self._allocate_thread_block(
                            free_resources,
                            requirements.threads_per_mpi_process)
                    free_resources -= allocation

            elif isinstance(requirements, MPINodesResReq):
                raise RuntimeError(
                        'Node-based MPI resource requirements are not'
                        ' supported yet. Please make an issue on GitHub.')

        except InsufficientResourcesAvailable:
            if not self._allocations and not virtual:
                # There are no other allocations and it's still not
                # enough. Just give it all and hope for the best.
                _logger.warning((
                        'Instance {} requires more resources than are'
                        ' available in total. Oversubscribing this'
                        ' instance.').format(instance))
                allocation = copy(self._all_resources)
            else:
                raise

        self._allocations[instance] = allocation
        return allocation

    def _allocate_thread_block(
            self, free_resources: Resources, threads: int) -> Resources:
        """Allocate resources for a group of threads.

        This chooses a set of <threads> cores on the same node. It
        returns the allocated resources; it doesn't update
        self._allocations or free_resources.

        Args:
            threads: Number of cores
            free_resources: Available resources to allocate from

        Returns:
            The allocated resources
        """
        for node in free_resources.nodes():
            if len(free_resources.cores[node]) >= threads:
                available_cores = sorted(free_resources.cores[node])
                to_reserve = set(available_cores[:threads])
                return Resources({node: to_reserve})
        raise InsufficientResourcesAvailable()



"""A simple runner for Python-only models.

Starting instances is out of scope for MUSCLE3, but is also very
useful for testing and prototyping. So we have a little bit of
support for it in this module.
"""
import logging
import multiprocessing as mp
import multiprocessing.connection as mpc
import sys
from typing import Callable, Dict, List, Tuple, cast

from ymmsl import Configuration, Identifier, Model, Reference

from libmuscle.util import generate_indices
from libmuscle.manager.manager import Manager


__all__ = ['run_simulation']


_logger = logging.getLogger(__name__)


Pipe = Tuple[mpc.Connection, mpc.Connection]


class MMPServerController:
    def __init__(
            self, process: mp.Process, control_pipe: Pipe,
            manager_location: str) -> None:
        """Create an MMPServerController.

        This class controls a manager running in a separate process.

        Args:
            process: The process the server is running in
            control_pipe: The control pipe for the server process
            manager_location: Network location of the manager
        """
        self._process = process
        self._control_pipe = control_pipe
        self.manager_location = manager_location

    def stop(self) -> None:
        """Stop the server process.

        Tells the server to stop listening, handle any running
        requests, then quit, after which this function returns.
        """
        self._control_pipe[0].send(True)
        self._control_pipe[0].close()
        self._process.join()


def manager_process(control_pipe: Pipe, configuration: Configuration) -> None:
    """Run function for a separate manager process.

    Args:
        pipe: The pipe through which to communicate with the parent.
        configuration: The configuration to run.
    """
    control_pipe[0].close()
    manager = Manager(configuration)
    control_pipe[1].send(manager.get_server_location())

    # wait for shutdown command
    control_pipe[1].recv()
    control_pipe[1].close()
    manager.stop()


def start_server_process(configuration: Configuration) -> MMPServerController:
    """Starts a manager in a separate process.

    Args:
        configuration: The configuration to run.

    Returns:
        A controller through which the manager can be shut down.
    """
    control_pipe = mp.Pipe()
    process = mp.Process(target=manager_process,
                         args=(control_pipe, configuration),
                         name='MuscleManager')
    process.start()
    control_pipe[1].close()
    # wait for start
    manager_location = control_pipe[0].recv()

    return MMPServerController(process, control_pipe, manager_location)


def implementation_process(
        instance_id: str, manager_location: str,
        implementation: Callable) -> None:
    prefix_tag = '--muscle-prefix='
    name_prefix = str()
    index_prefix = list()   # type: List[int]

    instance = Reference(instance_id)

    for i, arg in enumerate(sys.argv):
        if arg.startswith(prefix_tag):
            prefix_str = arg[len(prefix_tag):]
            name_prefix, index_prefix = _parse_prefix(prefix_str)

            name, index = _split_reference(instance)
            if len(name_prefix) > 0:
                name = Reference(name_prefix) + name
            index = index_prefix + index

            # replace it with the combined one
            sys.argv[i] = '--muscle-instance={}'.format(str(name + index))
            break
    else:
        sys.argv.append('--muscle-instance={}'.format(instance_id))

    for arg in sys.argv:
        if arg.startswith('--muscle-manager='):
            break
    else:
        sys.argv.append(f'--muscle-manager={manager_location}')

    with open(f'muscle3.{instance}.log', 'w') as log_file:
        # Redirect an already-configured standard logging setup
        # Logger.handlers and StreamHandler.stream are private, so this
        # is dangerous in theory. But this part of the logging code
        # hasn't changed in two decades, and we can use introspection to
        # avoid crashes, so in practice, it'll work.
        root_logger = logging.getLogger()
        if hasattr(root_logger, 'handlers'):
            for h in root_logger.handlers:
                if isinstance(h, logging.StreamHandler):
                    if hasattr(h, 'stream'):
                        if h.stream == sys.stderr:
                            h.stream = log_file
                        elif h.stream == sys.stdout:
                            h.stream = log_file

        sys.stderr = log_file
        sys.stdout = log_file

        # chain call
        try:
            implementation()
        except Exception:
            _logger.exception(
                    f'Component {instance} crashed, please check the log file'
                    ' for error messages')
            raise


def _parse_prefix(prefix: str) -> Tuple[str, List[int]]:
    """Parse a --muscle-prefix argument.

    This is like a Reference, but not quite, because the
    initial identifier may be omitted. That is, [1][2] is
    also a valid prefix.

    This parses an initial identifier, subsequent identifiers
    separated by periods, then a list of square-bracketed integers.

    Args:
        prefix: The prefix to parse.

    Returns:
        The identifier sequence and the list of ints.
    """
    def parse_identifier(prefix: str, i: int) -> Tuple[str, int]:
        name = str()
        while i < len(prefix) and prefix[i] not in '[.':
            name += prefix[i]
            i += 1
        return name, i

    def parse_number(prefix: str, i: int) -> Tuple[int, int]:
        number = str()
        while i < len(prefix) and prefix[i] in '0123456789':
            number += prefix[i]
            i += 1
        return int(number), i

    name = str()
    index = list()  # type: List[int]
    i = 0

    if i == len(prefix):
        return name, index

    idt, i = parse_identifier(prefix, i)
    name += idt

    while i < len(prefix) and prefix[i] == '.':
        name += '.'
        part, i = parse_identifier(prefix, i + 1)
        name += part

    while i < len(prefix) and prefix[i] == '[':
        nmb, i = parse_number(prefix, i + 1)
        index.append(nmb)
        if prefix[i] != ']':
            raise ValueError('Missing closing bracket in'
                             ' --muscle-prefix.')
        i += 1

    if i < len(prefix):
        raise ValueError(('Found invalid extra character {} in'
                          ' --muscle-prefix.').format(prefix[i]))

    return name, index


def _split_reference(ref: Reference) -> Tuple[Reference, List[int]]:
    index = list()     # type: List[int]
    i = 0
    while i < len(ref) and isinstance(ref[i], Identifier):
        i += 1
    name = ref[:i]

    while i < len(ref) and isinstance(ref[i], int):
        index.append(cast(int, ref[i]))
        i += 1

    return name, index


def run_instances(
        instances: Dict[str, Callable], manager_location: str) -> None:
    """Runs the given instances and waits for them to finish.

    The instances are described in a dictionary with their instance
    id (e.g. 'macro' or 'micro[12]' or 'my_mapper') as the key, and
    a function to run as the corresponding value. Each instance
    will be run in a separate process.

    Args:
        instances: A dictionary of instances to run
        manager_location: Network location of the manager
    """
    instance_processes = list()
    for instance_id_str, implementation in instances.items():
        instance_id = Reference(instance_id_str)
        process = mp.Process(
                target=implementation_process,
                args=(instance_id_str, manager_location, implementation),
                name='Instance-{}'.format(instance_id))
        process.start()
        instance_processes.append(process)

    failed_processes = list()
    for instance_process in instance_processes:
        instance_process.join()
        if instance_process.exitcode != 0:
            failed_processes.append(instance_process)

    if len(failed_processes) > 0:
        failed_names = map(lambda x: x.name, failed_processes)
        raise RuntimeError('Instances {} failed to shut down cleanly, please'
                           ' check the logs to see what went wrong.'.format(
                               ', '.join(failed_names)))


def run_simulation(
        configuration: Configuration, implementations: Dict[str, Callable]
        ) -> None:
    """Runs a simulation with the given configuration and instances.

    The yMMSL document must contain both a model and settings.

    This function will start the necessary instances described in
    the yMMSL document. To do so, it needs the corresponding
    implementations, which are given as a dictionary mapping the
    implementation name to a Python function (or any callable).

    Args:
        configuration: A description of the model and settings.
        instances: A dictionary of instances to run.
    """
    if not isinstance(configuration.model, Model):
        raise ValueError('The model description does not include a model'
                         ' definition, so the simulation can not be run.')

    configuration.model.check_consistent()

    instances = dict()
    for ce in configuration.model.components:
        impl_name = str(ce.implementation)
        if impl_name not in implementations:
            raise ValueError(('The model specifies an implementation named'
                              ' "{}" but the given set of implementations does'
                              ' not include it.').format(impl_name))

        impl_fn = implementations[impl_name]
        if not ce.multiplicity:
            instances[str(ce.name)] = impl_fn
        else:
            for index in generate_indices(ce.multiplicity):
                instance_id = str(ce.name + index)
                instances[instance_id] = impl_fn

    controller = start_server_process(configuration)
    try:
        run_instances(instances, controller.manager_location)
    finally:
        controller.stop()



from pathlib import Path
import logging
import sys
import traceback
from typing import Optional

from ymmsl import PartialConfiguration, save as save_ymmsl

from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPServer
from libmuscle.manager.instance_manager import InstanceManager
from libmuscle.manager.run_dir import RunDir
from libmuscle.manager.snapshot_registry import SnapshotRegistry
from libmuscle.manager.topology_store import TopologyStore


_logger = logging.getLogger(__name__)


class Manager:
    """The MUSCLE3 manager.

    This creates and manages instances and connects them together
    according to the simulation configuration.
    """
    def __init__(
            self, configuration: PartialConfiguration,
            run_dir: Optional[RunDir] = None, log_level: Optional[str] = None
            ) -> None:
        """Create a Manager.

        This creates the manager and the associated server, but does
        not start any instances.

        Args:
            configuration: The simulation configuration.
            run_dir: Main working directory.
        """
        self._configuration = configuration
        self._run_dir = run_dir
        log_dir = self._run_dir.path if self._run_dir else Path.cwd()
        self._logger = Logger(log_dir, log_level)
        self._topology_store = TopologyStore(configuration)
        self._instance_registry = InstanceRegistry()
        if run_dir is not None:
            snapshot_dir = run_dir.snapshot_dir()
        else:
            snapshot_dir = Path.cwd()
            if self._configuration.checkpoints:
                _logger.warning('Checkpoints are configured but no run'
                                ' directory is provided. Snapshots will be'
                                ' stored in the current working directory.')

        if self._run_dir:
            save_ymmsl(
                    self._configuration,
                    self._run_dir.path / 'configuration.ymmsl')

        self._instance_manager = None    # type: Optional[InstanceManager]
        try:
            configuration = self._configuration.as_configuration()
            if self._run_dir is not None:
                self._instance_manager = InstanceManager(
                        configuration, self._run_dir)
        except ValueError:
            pass

        # SnapshotRegistry creates a worker thread, must be created after
        # instance_manager which forks the process
        self._snapshot_registry = SnapshotRegistry(
                configuration, snapshot_dir, self._topology_store)
        self._snapshot_registry.start()

        self._server = MMPServer(
                self._logger, self._configuration,
                self._instance_registry, self._topology_store,
                self._snapshot_registry, run_dir)

        if self._instance_manager:
            self._instance_manager.set_manager_location(
                    self.get_server_location())

    def get_server_location(self) -> str:
        """Returns the network location of the server."""
        return self._server.get_location()

    def start_instances(self) -> None:
        """Starts all required component instances."""
        if self._run_dir is None:
            raise RuntimeError('No run dir specified')
        if not self._instance_manager:
            raise RuntimeError(
                    'For MUSCLE3 to be able to start instances, the'
                    ' configuration must contain a model, implementations,'
                    ' and resources. Please make sure they are all there.')
        try:
            self._instance_manager.start_all()
        except:     # noqa
            _logger.error('An error occurred while starting the components:')
            for line in traceback.format_exception(*sys.exc_info()):
                _logger.error(line)
            self._instance_manager.shutdown()
            raise

    def stop(self) -> None:
        """Shuts down the manager."""
        # self._server.stop()
        self._server.stop()
        self._snapshot_registry.shutdown()
        self._snapshot_registry.join()
        self._logger.close()

    def wait(self) -> bool:
        """Blocks until the simulation is done, then shuts down.

        Returns:
            True if success, False if an error occurred.
        """
        if self._instance_manager:
            try:
                success = self._instance_manager.wait()
            finally:
                self._instance_manager.shutdown()
        else:
            self._instance_registry.wait()
            success = True
        self.stop()
        return success



from libmuscle.outbox import Outbox
from libmuscle.mpp_message import MPPMessage

from copy import copy
import pytest

from ymmsl import Reference


@pytest.fixture
def outbox():
    return Outbox()


@pytest.fixture
def message():
    Ref = Reference
    return MPPMessage(
            Ref('sender.out'), Ref('receiver.in'),
            None, 0.0, 1.0,
            bytes(),
            0, 1.0,
            'testing'.encode('utf-8'))


def test_create_outbox():
    box = Outbox()
    assert box._Outbox__queue.qsize() == 0


def test_deposit_message(outbox, message):
    outbox.deposit(message)
    assert outbox._Outbox__queue.qsize() == 1
    assert outbox._Outbox__queue.get(message)


def test_retrieve_message(outbox, message):
    outbox._Outbox__queue.put(message)
    assert outbox.retrieve() == message


def test_deposit_retrieve_order(outbox, message):
    m1 = copy(message)
    m2 = copy(message)

    outbox.deposit(m1)
    outbox.deposit(m2)

    assert outbox.retrieve() == m1
    assert outbox.retrieve() == m2



import socket


def test_create(tcp_transport_server):
    assert tcp_transport_server._server_thread.is_alive()


def test_location(tcp_transport_server):
    assert tcp_transport_server.get_location().startswith('tcp:')


def test_request(tcp_transport_server):
    request = b'testing'
    response = b'response'

    tcp_transport_server._handler.handle_request.return_value = response

    location = tcp_transport_server._server.server_address
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.connect(location)
        length = len(request).to_bytes(8, byteorder='little')
        sock.sendall(length)
        sock.sendall(request)

        lenbuf = bytearray(8)
        sock.recv_into(lenbuf, 8)
        length = int.from_bytes(lenbuf, 'little')
        assert length == len(response)

        databuf = bytearray(length)
        received_count = 0
        while received_count < length:
            bytes_left = length - received_count
            received_count += sock.recv_into(
                    memoryview(databuf)[received_count:], bytes_left)

    assert databuf == response
